unit WdxFieldsProc;
//    Super_WDX

interface

uses
  Windows, SysUtils, ContPlug, Generics.Collections;

type
  TContentGetSupportedField = function(FieldIndex: integer; FieldName: PAnsiChar;
    Units: PAnsiChar; maxlen: integer): integer; stdcall;
  TContentGetValue = function(FileName: PAnsiChar; FieldIndex, UnitIndex: integer;
    FieldValue: pbyte; maxlen, flags: integer): integer; stdcall;
  TContentGetValueW = function(FileName: PWideChar; FieldIndex, UnitIndex: integer;
    FieldValue: pbyte; maxlen, flags: integer): integer; stdcall;
  TContentSetDefaultParams = procedure
    (dps: pContentDefaultParamStruct); stdcall;
  TContentUnloading = procedure(); stdcall;
  TContentGetDetectString = procedure(DetectString: PAnsiChar; MaxLen: Integer); stdcall;
  TContentGetFormattedValueW = function(FileName, ContentString: PChar;
    FieldValue: PChar; maxlen: integer): integer; stdcall;

type
  TPluginInfo = packed record
    Name: string;
    Handle: THandle;
    NotValid: Boolean;
    SelfLoaded: Boolean;
    //ContentGetValue: TContentGetValue;
    //ContentGetValueW: TContentGetValueW;
    //Units, Fields: array of string;
  end;

function GetWdxFieldName(const fPlugin: string; FieldNum: integer): string;
function GetWdxFieldNum(const fPlugin: string; FieldName: string;
  UnitName: string; var UnitNum: byte): integer;
function GetWdxField(const fPlugin, fFile: string;
  const FieldNumber, UnitNum: byte): string;
procedure DebugLog(LogMessage: string);
procedure SelfUnloading;
function GetFormattedValue(PluginName, FileName, ContentString: string): string;

var
  PluginPath: string;
  // -   -   ,      ,    
  ShowError: boolean; //      
  Debug: boolean; //  
  EmptyReplace: boolean; //      .
  CustomFoldersView: integer; //      
  CustomFolderFileName: string = 'folder.ini'; //    
  DateTimeISO: Boolean = False;
  TC_ContentIni: AnsiString;
  AllowUnloadPlugins: Boolean = True;

  GF: byte;

implementation

var
  Plugins: TDictionary<string, TPluginInfo>;

procedure DebugLog(LogMessage: string); //    
var
  f: textfile;
begin
  if not Debug then
    exit;
  if not fileexists(IncludeTrailingPathDelimiter(extractfilepath(PluginPath)) +
    'DebugLog.txt') then
  begin
    AssignFile(f, IncludeTrailingPathDelimiter(extractfilepath(PluginPath)) +
      'DebugLog.txt');
    rewrite(f);
    closefile(f);
  end;
  AssignFile(f, IncludeTrailingPathDelimiter(extractfilepath(PluginPath)) +
    'DebugLog.txt');
  Append(f);
  Writeln(f, datetostr(now) + '-' + timetostr(now) + ': ' + LogMessage);
  closefile(f);
end;

function WdxFieldType(n: integer; ShowError: Boolean): string;
begin
  if not ShowError then
    result := ''
  else
    case n of
      FT_NUMERIC_32:
        result := 'FT_NUMERIC_32';
      FT_NUMERIC_64:
        result := 'FT_NUMERIC_64';
      FT_NUMERIC_FLOATING:
        result := 'FT_NUMERIC_FLOATING';
      FT_DATE:
        result := 'FT_DATE';
      FT_TIME:
        result := 'FT_TIME';
      FT_DATETIME:
        result := 'FT_DATETIME';
      FT_BOOLEAN:
        result := 'FT_BOOLEAN';
      FT_MULTIPLECHOICE:
        result := 'FT_MULTIPLECHOICE';
      FT_STRING:
        result := 'FT_STRING';
      FT_FULLTEXT:
        result := 'FT_FULLTEXT';
      FT_NOSUCHFIELD:
        result := 'FT_NOSUCHFIELD';
      FT_FILEERROR:
        result := 'FT_FILEERROR';
      FT_FIELDEMPTY:
        result := 'FT_FIELDEMPTY';
      FT_DELAYED:
        result := 'FT_DELAYED';
      ft_StringW:
        result := 'FT_STRINGW';
      FT_FULLTEXTW:
        result := 'FT_FULLTEXTW';
    else
      result := Format('? (%d)', [n]);
    end;
end;

// -----------------------------------------------
var
  fieldsNum: integer;

procedure UnloadPlugin(var PluginHandle: THandle; Force: Boolean = False);
var
  ContentUnloading: TContentUnloading;
begin
  if not Force then
    Exit;
  if PluginHandle <= 0 then Exit;
  try

    @ContentUnloading := GetProcAddress(PluginHandle, 'ContentPluginUnloading');
    try
      if Assigned(ContentUnloading) then
        ContentUnloading;
    finally
      FreeLibrary(PluginHandle);
      PluginHandle := 0;
    end;

  except
  end;
end;

function LoadPlugin(const PluginName: string): THandle;
var
  dps: TContentDefaultParamStruct;
  Proc3: TContentSetDefaultParams;
  ProcGetDetectString: TContentGetDetectString;
  Buf: PAnsiChar;
  PluginInfo: TPluginInfo;
begin
  Result := 0;

  //result := GetModuleHandle(PluginName);
  //         .
  //if result = 0 then //   

  if Plugins.TryGetValue(AnsiLowerCase(PluginName), PluginInfo) then
  begin
    if PluginInfo.NotValid then
      Result := 0
    else
      Result := PluginInfo.Handle;
    Exit;
  end;

  if not FileExists(PluginName) then
    Exit;

  PluginInfo.Name := AnsiLowerCase(PluginName);
  PluginInfo.Handle := 0;
  PluginInfo.SelfLoaded := True;

  Result := SafeLoadLibrary(PluginName);

  if Result = 0 then
  begin
    Plugins.Add(AnsiLowerCase(PluginName), PluginInfo);
    Exit;
  end;

  PluginInfo.Handle := Result;

  try
    try
      //   ContentSetDefaultParams  
      @Proc3 := GetProcAddress(Result, 'ContentSetDefaultParams');
      if @Proc3 <> nil then
      begin
        FillChar(dps, SizeOf(dps), 0);
        dps.Size := SizeOf(dps);
        dps.PluginInterfaceVersionLow := 30;
        dps.PluginInterfaceVersionHi := 1;
        //StrLCopy(dps.DefaultIniName, PAnsiChar(TC_ContentIni), Length(dps.DefaultIniName) - 1);
        StrLCopy(dps.DefaultIniName,
          PAnsiChar(AnsiString(ExtractFilePath(PluginPath)) + 'Plugins.ini'),
          Length(dps.DefaultIniName) - 1);
        Proc3(@dps);
      end;

      @ProcGetDetectString := GetProcAddress(Result, 'ContentGetDetectString');
      if @ProcGetDetectString <> nil then
      begin
        GetMem(Buf, 2048);
        try
          ProcGetDetectString(Buf, 2046);
        finally
          FreeMem(Buf);
        end;
      end;

    finally
      Plugins.Add(AnsiLowerCase(PluginName), PluginInfo);
    end;

  except
    //UnloadPlugin(Result);
    PluginInfo.NotValid := True;
    Plugins.AddOrSetValue(AnsiLowerCase(PluginName), PluginInfo);
    Result := 0;
  end;
end;

procedure SelfUnloading;
var
  Pair: TPair<string, TPluginInfo>;
  PluginInfo: TPluginInfo;
begin
  try
    if AllowUnloadPlugins then
      for Pair in Plugins do
      begin
        PluginInfo.Handle := Pair.Value.Handle;
        if PluginInfo.SelfLoaded then
        try
          UnloadPlugin(PluginInfo.Handle, True);
        except
        end;
    end;
  finally
    Plugins.Clear;
  end;
end;

{ -------function not used in fdx plugin, only in ssetings---------------------- }
//       .
function GetWdxFieldName(const fPlugin: string; FieldNum: integer): string;
var
  hLib: THandle;
  Proc1: TContentGetSupportedField;
  buf1, buf2: array [0 .. 2 * 1024] of AnsiChar;
  res: integer;
begin
  result := '';
  res := -1;
  hLib := LoadPlugin(fPlugin); //   
  if hLib = 0 then
    exit; //   .      if (hLib=0) or (hLib=INVALID_HANDLE_VALUE,   . :     -    ,       (   )
  @Proc1 := GetProcAddress(hLib, 'ContentGetSupportedField');
  //    ContentGetSupportedField  
  if @Proc1 = nil then
  begin
    UnloadPlugin(hLib);
    exit
  end; //   ,     .
  try
    FillChar(buf1, SizeOf(buf1), 0); //   
    FillChar(buf2, SizeOf(buf2), 0);
    res := Proc1(FieldNum + 1, buf1, buf2, SizeOf(buf1));
    //  ContentGetSupportedField,  .
    // res -   (.    ), buf1 -  fieldNum+1  .
    if res = ft_nomorefields then
    //   ft_nomorefields -    .
    begin
      result := '';
    end;
    result := buf1;
    UnloadPlugin(hLib);
  except
    on E: exception do
    begin
      DebugLog('Some malfunction in procedure GetWdxFieldName.' + #10 + #13 +
        '  GetWdxFieldName.' + #10 + #13 + 'Variable values:' + #10 +
        #13 + 'fPlugin: ' + fPlugin + #10 + #13 + 'FieldNum: ' +
        inttostr(FieldNum) + #10 + #13 + 'buf1: ' + buf1 + #10 + #13 + 'buf2: '
        + buf2 + #10 + #13 + 'res: ' + inttostr(res) + #10 + #13 + 'hLib: ' +
        inttostr(hLib));
    end;
  end;
end;

// ,   -      
//    1.0   (UnitName -  ,  UnitNum   ).
function GetWdxFieldNum(const fPlugin: string; FieldName: string;
  UnitName: string; var UnitNum: byte): integer;
var
  hLib: THandle;
  Proc1: TContentGetSupportedField;
  buf1, buf2: array [0 .. 2 * 1024] of AnsiChar;
  res: integer;
  i: byte;
  tmpbuf, tmpunit: string;
begin
  i := 0;
  result := -1;
  res := -1;
  hLib := LoadPlugin(fPlugin); //  
  if hLib = 0 then
    exit;
  try
    @Proc1 := GetProcAddress(hLib, 'ContentGetSupportedField');
    if @Proc1 = nil then
    begin
      UnloadPlugin(hLib);
      exit
    end;
    fieldsNum := -1;
    repeat
      FillChar(buf1, SizeOf(buf1), 0);
      FillChar(buf2, SizeOf(buf2), 0);
      res := Proc1(fieldsNum + 1, buf1, buf2, SizeOf(buf1));
      //     ,     ft_nomorefields (  )      .     while,  
      if res = ft_nomorefields then
        Break; //   
      Inc(fieldsNum);
      if AnsiSameText(string(AnsiString(buf1)), FieldName) then
      //        
      begin
        result := fieldsNum + 1;
        //   -   (+1       1 )
        if buf2[0] <> #0 then //    
        begin
          if UnitName <> '' then
          begin
            tmpbuf := buf2;
            tmpbuf := tmpbuf + '|';
            repeat
              tmpunit := copy(tmpbuf, 1, pos('|', tmpbuf) - 1);
              delete(tmpbuf, 1, pos('|', tmpbuf));
              Inc(i);
            until (lowercase(tmpunit) = lowercase(UnitName)) or
              (Length(tmpbuf) = 0);
            UnitNum := i - 1;
          end
          else
            UnitNum := 0;
        end;
        Break; //   
      end;
    until False;
    UnloadPlugin(hLib);
  except
    on E: exception do
    begin
      DebugLog('Some malfunction in procedure GetWdxFieldNum.' + #10 + #13 +
        '  GetWdxFieldNum.' + #10 + #13 + 'Variable values:' + #10 +
        #13 + 'fPlugin: ' + fPlugin + #10 + #13 + 'FieldName: ' + FieldName +
        #10 + #13 + 'buf1: ' + buf1 + #10 + #13 + 'buf2: ' + buf2 + #10 + #13 +
        'res: ' + inttostr(res) + #10 + #13 + 'hLib: ' + inttostr(hLib));
    end;
  end;
end;

//    fPlugin    FieldNumber   fFile.
//    1.0  
function GetWdxField(const fPlugin, fFile: string;
  const FieldNumber, UnitNum: byte): string;
const
  BufLen = 2048;
var
  hLib: THandle;
  Proc1: TContentGetValue;
  Proc2: TContentGetValueW;
  //buf1, buf2: array [0 .. 2 * 1024] of Byte;
  buf1: PByte;
  stime: TSystemTime;
  sval, C: string;
  A: AnsiString;
  res: Integer;
  NumericFloatTextW: PChar;
  NumericFloatTextA: PAnsiChar;
  IsWideFunction: Boolean;

  function GetValue: Boolean;
  begin
    Result := False;
    IsWideFunction := False;
    @Proc2 := GetProcAddress(hLib, 'ContentGetValueW');
    if @Proc2 <> nil then
    begin
      {  ContentGetValue,   ,  ,
       UnitIndex    -  0,     
        . Res -   }
      res := Proc2(PWideChar(string(fFile)), FieldNumber - 1,
        UnitNum, buf1, BufLen - 2, 0);
      Result := True;
      IsWideFunction := True;
    end
    { $IFNDEF WIN64}
    else begin
      @Proc1 := GetProcAddress(hLib, 'ContentGetValue');
      if @Proc1 <> nil then
      begin
        res := Proc1(PAnsiChar(AnsiString(fFile)), FieldNumber - 1,
          UnitNum, buf1, BufLen - 2, 0);
        Result := True;
      end
    end;
    { $ENDIF}
  end;

begin
  Result := '';
  hLib := 0;
  //     UnitNum,     :(
  hLib := LoadPlugin(fPlugin); //   
  if hLib = 0 then
    exit;

  buf1 := nil;
  try
    Res := 0;

    try
      buf1 := AllocMem(BufLen);

      if not GetValue then
        Exit;

      fieldsNum := FieldNumber;
      sval := '';
      case res of //        String.
        FT_FIELDEMPTY:
          sval := '';
        FT_NUMERIC_32:
          sval := IntToStr(PInteger(Buf1)^);
        FT_NUMERIC_64:
          sval := IntToStr(PInt64(Buf1)^);
        FT_NUMERIC_FLOATING:
          begin
            sval := FormatFloat('0.000', PDouble(Buf1)^);
            //sval := FloatToStr(PDouble(Buf1)^);

            C := '';
            if IsWideFunction then
            begin
              NumericFloatTextW := PChar(buf1 + SizeOf(Double));
              C := NumericFloatTextW;
            end else begin
              NumericFloatTextA := PAnsiChar(buf1 + SizeOf(Double));
              A := NumericFloatTextA;
              C := string(A);
            end;
            if Length(C) <> 0 then
              sval := Format('%s (%s)', [C, sval]);
          end;
        FT_DATE:
          sval := Format('%2.2d.%2.2d.%4.4d', [PDateFormat(buf1)^.wDay,
            PDateFormat(buf1)^.wMonth, PDateFormat(buf1)^.wYear]);
        FT_TIME:
          sval := Format('%2.2d:%2.2d:%2.2d', [PTimeFormat(buf1)^.wHour,
            PTimeFormat(buf1)^.wMinute, PTimeFormat(buf1)^.wSecond]);

        FT_DATETIME:
          begin
            FileTimeToSystemTime(PFileTime(buf1)^, stime);
            if DateTimeISO then
              sval := Format('%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d',
                [stime.wYear, stime.wMonth, stime.wDay, stime.wHour,
                stime.wMinute, stime.wSecond])
            else
              sval := Format('%2.2d.%2.2d.%4.4d %2.2d:%2.2d:%2.2d',
                [stime.wDay, stime.wMonth, stime.wYear, stime.wHour,
                stime.wMinute, stime.wSecond]);
          end;

        FT_BOOLEAN:
          if PBoolean(buf1)^ then
            sval := 'TRUE'
          else
            sval := 'FALSE';

        FT_STRING, FT_MULTIPLECHOICE, FT_FULLTEXT:
          sval := PAnsiChar(buf1);

        ft_StringW, ft_FullTextW:
          sval := PChar(buf1);
      else
        sval := WdxFieldType(res, ShowError);
        //     -     .
      end;

      Result := sval;
    except
      on E: exception do
      begin
        DebugLog('Some malfunction in procedure GetWdxField.' + #10 + #13 +
          '  GetWdxField.' + #10 + #13 + 'Variable values:' + #10 + #13
          + 'fPlugin: ' + fPlugin + #10 + #13 + 'fFile: ' + fFile + #10 + #13 +
          'FieldNumber: ' + inttostr(FieldNumber) + #10 + #13 +
          'dps.DefaultIniName: ' + ExtractFilePath(PluginPath) +  'Plugins.ini' +
          #10 + #13 + 'buf1: ' +
          sval + #10 + #13 + 'res: ' + inttostr(res)
          + #10 + #13 + 'sval: ' + sval + #10 + #13 + 'hLib: ' + inttostr(hLib));
        result := 'ERROR!';
        result := E.Message;
        //      try  .
      end;
    end;

  finally
    FreeMem(buf1);
    UnloadPlugin(hLib);
  end;

end;

function GetFormattedValue(PluginName, FileName, ContentString: string): string;
var
  hLib: THandle;
  GFV: TContentGetFormattedValueW;
  res: array [0 .. 2 * 1024] of Char;
begin
  result := '';
  hLib := LoadPlugin(PluginName); //   
  if hLib = 0 then
    exit;
  @GFV := GetProcAddress(hLib, 'GetFormattedValueW');
  if @GFV <> nil then
  begin
    FillChar(res, SizeOf(res), 0);
    if GFV(PChar(FileName), PChar(ContentString), @res, SizeOf(res)) = 1 then
      result := res
    else
      result := 'Can''t get value';
  end;
end;

initialization
  Plugins := TDictionary<string, TPluginInfo>.Create;

finalization
  SelfUnloading;
  FreeAndNil(Plugins);

end.


// That`s all.
